#!/usr/bin/env python2
# coding:utf-8

import os
import sys
import time
import signal
import argparse
import logging
import functools

from subprocess import call, PIPE
from threading import Thread
from datetime import datetime
from collections import Counter

from qsub import qsub
from version import __version__


class ParseSingal(Thread):
    def __init__(self):
        super(ParseSingal, self).__init__()
        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)

    def run(self):
        time.sleep(1)

    def signal_handler(self, signum, frame):
        stillrunjob = qjobs.jobqueue.queue
        if clear:
            pid = os.getpid()
            gid = os.getpgid(pid)
            for jn in stillrunjob:
                qjobs.state[jn] = "kill"
                qjobs.error.add(jn)
            sumJobs(qjobs)
            call('qdel "*_%d"' % os.getpid(),
                 shell=True, stderr=PIPE, stdout=PIPE)
            call("kill -9 -%d" % gid, shell=True, stderr=PIPE, stdout=PIPE)
        else:
            for jn in stillrunjob:
                qjobs.state[jn] = "run"
            sumJobs(qjobs)
        sys.exit(signum)


def LogExc(f):
    @functools.wraps(f)
    def wrapper(*largs, **kwargs):
        try:
            res = f(*largs, **kwargs)
        except Exception, e:
            logging.getLogger(__name__).exception(e)
            os.kill(os.getpid(), 15)
        return res
    return wrapper


def parseArgs():
    parser = argparse.ArgumentParser(
        description="For manger submit your jobs in a job file.")
    parser.add_argument("-n", "--num", type=int,
                        help="the max job number runing at the same time, default: 1000", default=1000, metavar="<int>")
    parser.add_argument("-j", "--jobfile", type=str, required=True,
                        help="the input jobfile", metavar="<jobfile>")
    parser.add_argument('-i', '--injname', help="job names you need to run, default: all jobnames of you job file",
                        nargs="*", type=str, metavar="<str>")
    parser.add_argument('-s', '--start', help="job beginning with the number you given, 1 by default",
                        type=int, default=1, metavar="<int>")
    parser.add_argument('-e', '--end', help="job ending with the number you given, last job by default",
                        type=int, metavar="<int>")
    parser.add_argument('-r', '--resub', help="rebsub you job when error, 0 or minus means do not re-submit, 3 times by default",
                        type=int, default=3, metavar="<int>")
    parser.add_argument('-ivs', '--resubivs', help="rebsub interval seconds, 2 by default",
                        type=int, default=2, metavar="<int>")
    parser.add_argument("-m", '--mode', type=str, default="sge", choices=[
                        "sge", "localhost"], help="the mode to submit your jobs, 'sge' by default, if no sge installed, always localhost.")
    parser.add_argument("-nc", '--noclean', action="store_false", help="whether to clean all jobs or subprocess created by this programe when the main process exits, default: clean.",
                        default=True)
    parser.add_argument("--strict", action="store_true", default=False,
                        help="use strict to run. Means if any errors occur, clean all jobs and exit programe. off by default")
    parser.add_argument("-l", "--log", type=str,
                        help='append log info to file, sys.stdout by default', metavar="<file>")
    parser.add_argument('-v', '--version',
                        action='version', version="%(prog)s v" + __version__)
    return parser.parse_args()


def Mylog(logfile=None, level="info", name=None):
    logger = logging.getLogger(name)
    f = logging.Formatter(
        '[%(levelname)s %(asctime)s] %(message)s')
    if logfile is None:
        h = logging.StreamHandler(sys.stdout)  # default: sys.stderr
    else:
        h = logging.FileHandler(logfile, mode='w')
    h.setFormatter(f)
    logger.addHandler(h)
    if level.lower() == "info":
        logger.setLevel(logging.INFO)
    elif level.lower() == "debug":
        logger.setLevel(logging.DEBUG)
    return logger


def sumJobs(qjobs):
    thisrunjobs = set([j.name for j in qjobs.jobs])
    realrunjobs = thisrunjobs - qjobs.has_success
    realrunsuccess = qjobs.success - qjobs.has_success
    realrunerror = qjobs.error
    resubjobs = set(
        [k for k, v in qjobs.subtimes.items() if v != qjobs.times])
    thisjobstates = qjobs.state
    # qjobs.writejob(qjobs.jfile + ".bak")  ## write a new job file

    logger = logging.getLogger(__name__)
    if len(realrunerror) == 0:
        logger.info("All tesks(total(%d), actual(%d), actual_success(%d), actual_error(%d)) in file (%s) finished successfully.",
                    len(thisrunjobs), len(realrunjobs), len(realrunsuccess), len(realrunerror), os.path.abspath(qjobs.jfile))
    else:
        logger.info("All tesks( total(%d), actual(%d), actual_success(%d), actual_error(%d) ) in file (%s) finished, But there are ERROR tesks.",
                    len(thisrunjobs), len(realrunjobs), len(realrunsuccess), len(realrunerror), os.path.abspath(qjobs.jfile))
    if len(qjobs.jobs) == len(qjobs.totaljobs):
        qjobs.writestates(os.path.join(
            qjobs.logdir, "job.status.txt"))
    logger.info(str(dict(Counter(thisjobstates.values()))))


@LogExc
def main():
    args = parseArgs()
    logger = Mylog(logfile=args.log, name=__name__)
    global clear, qjobs
    clear = args.noclean
    h = ParseSingal()
    h.start()
    qjobs = qsub(args.jobfile, args.num, args.injname,
                 args.start, args.end, mode=args.mode, usestrict=args.strict)
    qjobs.run(times=args.resub - 1, resubivs=args.resubivs)
    sumJobs(qjobs)


if __name__ == "__main__":
    main()



#!/usr/bin/env python2
# coding:utf-8

import os
import re
import sys
from commands import getstatusoutput


class Jobfile(object):
    def __init__(self, jobfile, mode=None):
        self.has_sge = True if getstatusoutput(
            'command -v qstat')[0] == 0 else False

        self._path = os.path.abspath(jobfile)
        if not os.path.exists(self._path):
            raise IOError("No such file: %s" % self._path)
        self._pathdir = os.path.dirname(self._path)
        self.logdir = os.path.join(self._pathdir, "log")
        if self.has_sge:
            self.mode = "sge" if mode is None else mode
        else:
            self.mode = "localhost"

    def orders(self):
        orders = {}
        with open(self._path) as fi:
            for line in fi:
                if not line.strip() or line.startswith("#"):
                    continue
                if line.startswith("order"):
                    line = line.split()
                    if "after" in line:
                        idx = line.index("after")
                        o1 = line[1:idx]
                        o2 = line[idx+1:]
                    elif "before" in line:
                        idx = line.index("before")
                        o1 = line[idx+1:]
                        o2 = line[1:idx]
                    for o in o1:
                        if o in self.alljobnames:
                            for i in o2:
                                if i not in self.alljobnames:
                                    self.throw("order names (%s) not in job, (%s)" % (
                                        i, " ".join(line)))
                                else:
                                    orders.setdefault(o, set()).add(i)
                        else:
                            self.throw("order names (%s) not in job, (%s)" % (
                                o, " ".join(line)))
        return orders

    def jobs(self, names=None, start=1, end=None):  # real this jobs, not total jobs
        jobs = []
        job = []
        with open(self._path) as fi:
            for line in fi:
                if not line.strip() or line.strip().startswith("#"):
                    continue
                if "#" in line:
                    line = line[line.index("#"):]
                line = line.strip()
                if line.startswith("log_dir"):
                    self.logdir = os.path.normpath(
                        os.path.join(self._pathdir, line.split()[-1]))
                    continue
                if line == "job_begin":
                    if len(job) and job[-1] == "job_end":
                        jobs.append(Job(self, job))
                        job = [line, ]
                    else:
                        job.append(line)
                elif line.startswith("order"):
                    continue
                else:
                    job.append(line)
            if len(job):
                jobs.append(Job(self, job))
        self.totaljobs = jobs
        self.alljobnames = [j.name for j in jobs]
        thisjobs = []
        if names is not None:
            for jn in jobs:
                name = jn.name
                for namereg in names:
                    if re.search(namereg, name):
                        thisjobs.append(jn)
                        break
            return thisjobs
        jobend = len(jobs) if end is None else end
        thisjobs = self.totaljobs[start-1:jobend]
        return thisjobs

    def throw(self, msg):
        raise OrderError(msg)


class Job(object):
    def __init__(self, jobfile, rules):
        self.rules = rules
        self.jf = jobfile
        self.name = None
        self.status = None
        self.sched_options = None
        self.cmd = []
        self.host = None
        self.checkrule()
        cmd = False
        for j in self.rules:
            j = j.strip()
            if not j or j.startswith("#"):
                continue
            if j in ["job_begin", "job_end"]:
                continue
            elif j.startswith("name"):
                name = j.split()[1:]
                if len(name) > 1:
                    self.name = self.name.replace(" ", "_")
                else:
                    self.name = name[-1]
            elif j.startswith("status"):
                self.status = j.split()[-1]
            elif j.startswith("sched_options"):
                self.sched_options = " ".join(j.split()[1:])
            elif j.startswith("host"):
                self.host = j.split()[-1]
            elif j == "cmd_begin":
                cmd = True
                continue
            elif j == "cmd_end":
                cmd = False
            elif j.startswith("cmd"):
                self.cmd.append(" ".join(j.split()[1:]))
            elif j.startswith("memory"):
                self.sched_options += " -l h_vmem=" + j.split()[-1].upper()
            elif j.startswith("time"):
                pass  # miss
                # self.sched_options += " -l h_rt=" + j.split()[-1].upper()   # hh:mm:ss
            else:
                if cmd:
                    self.cmd.append(j)
                else:
                    self.throw("cmd after cmd_end")
        for c in self.cmd:
            if c.startswith("exit"):
                self.throw(
                    "'exit' command not allow in the cmd string in %s job." % self.name)
        if self.jf.has_sge:
            # if localhost defined, run localhost whether sge installed.
            if self.jf.mode == "localhost":
                self.host = "localhost"
            else:
                # if self.host has been defined other mode in job, sge default.
                if self.host not in [None, "sge", "localhost"]:
                    self.host = None
        else:
            self.host = "localhost"  # if not sge installed, only run localhost
        if len(self.cmd) > 1:
            self.cmd = " && ".join(self.cmd)
        elif len(self.cmd) == 1:
            self.cmd = self.cmd[0]
        else:
            self.throw("No cmd in %s job" % self.name)

    def checkrule(self):
        rules = self.rules[:]
        if len(rules) <= 4:
            self.throw("rules lack of elements")
        if rules[0] != "job_begin" or rules[-1] != "job_end":
            self.throw("No start or end in you rule")
        rules = rules[1:-1]
        if any([i.startswith("cmd") for i in rules]):
            return
        try:
            rules.remove("cmd_begin")
            rules.remove("cmd_end")
        except ValueError:
            self.throw("No start or end in %s job" % "\n".join(rules))

    def write(self, fo):
        fo.write("job_begin\n")
        rules = self.rules
        host = None
        for r in rules:
            r = r.strip()
            if r.startswith("host"):
                host = r.split()[-1]
        for k, attr in self.__dict__.items():
            if hasattr(attr, '__call__'):
                continue
            if k == "cmd":
                fo.write("    cmd_begin\n")
                cmdlines = ["        " +
                            i.strip() + "\n" for i in attr.split("&&")]
                fo.writelines(cmdlines)
                fo.write("    cmd_end\n")
            elif k == "rules":
                continue
            elif k == "host":
                if host is None:
                    continue
                else:
                    fo.write("    host " + host + "\n")
            else:
                try:
                    line = "    " + k + " " + attr + "\n"
                except TypeError:
                    continue
                fo.write(line)
        fo.write("job_end\n")

    def throw(self, msg):
        raise RuleError(msg)


class RuleError(Exception):
    def __init__(self, ErrorInfo):
        super(RuleError, self).__init__(self)
        self.errorinfo = ErrorInfo

    def __str__(self):
        return self.errorinfo
    __repr__ = __str__


class OrderError(Exception):
    def __init__(self, ErrorInfo):
        super(OrderError, self).__init__(self)
        self.errorinfo = ErrorInfo

    def __str__(self):
        return self.errorinfo
    __repr__ = __str__



#!/usr/bin/env python
# coding:utf-8
# 投递任务，将任务散播到集群SGE,并等待所有任务执行完成时退出。 杀掉本主进程时，由本进程投递的qsub任务也会被杀掉，不接受`kill -9`信号

import sys
import os
import time
import argparse
import signal
from subprocess import call, PIPE
from threading import Thread
from Queue import Queue
from datetime import datetime
from shutil import rmtree

RUNSTAT = " && echo [\`date +'%F %T'\`] SUCCESS || echo [\`date +'%F %T'\`] ERROR"


class ParseSingal(Thread):
    def __init__(self):
        super(ParseSingal, self).__init__()
        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)

    def run(self):
        time.sleep(1)

    def signal_handler(self, signum, frame):
        call('qdel "%s*"' % jn, shell=True, stderr=PIPE, stdout=PIPE)
        sys.exit(signum)


def not_empty(s):
    return s and s.strip() and (not s.strip().startswith("#")) and s.strip().lower() != "wait"


def parserArg():
    pid = os.getpid()
    parser = argparse.ArgumentParser(
        description="For qsub your jobs in a job file.")
    parser.add_argument("-q", "--queue", type=str, help="the queue your job running, default: all.q",
                        default="all.q", metavar="<queue>")
    parser.add_argument("-m", "--memory", type=int,
                        help="the memory used per command (GB), default: 1", default=1, metavar="<int>")
    parser.add_argument("-c", "--cpu", type=int,
                        help="the cpu numbers you job used, default: 1", default=1, metavar="<int>")
    parser.add_argument("-wd", "--workdir", type=str, help="work dir, default: %s" %
                        os.path.abspath(os.getcwd()), default=os.path.abspath(os.getcwd()), metavar="<workdir>")
    parser.add_argument("-N", "--jobname", type=str,
                        help="job name", metavar="<jobname>")
    parser.add_argument("-o", "--logdir", type=str,
                        help='the output log dir, default: "qsub.out.*"', metavar="<logdir>")
    parser.add_argument("-n", "--num", type=int,
                        help="the max job number runing at the same time. default: all in your job file", metavar="<int>")
    parser.add_argument("-s", "--startline", type=int,
                        help="which line number be used for the first job tesk. default: 1", metavar="<int>", default=1)
    parser.add_argument("-e", "--endline", type=int,
                        help="which line number (include) be used for the last job tesk. default: all in your job file", metavar="<int>")
    parser.add_argument("-b", "--block", action="store_true", default=False,
                        help="if passed, block when submitted you job, default: off")
    parser.add_argument("jobfile", type=str,
                        help="the input jobfile", metavar="<jobfile>")
    progargs = parser.parse_args()
    if progargs.logdir is None:
        progargs.logdir = os.path.join(os.path.abspath(os.path.dirname(
            progargs.jobfile)), "qsub.out."+os.path.basename(progargs.jobfile))
    if progargs.jobname is None:
        progargs.jobname = os.path.basename(progargs.jobfile) + "_" + str(pid) if not os.path.basename(
            progargs.jobfile)[0].isdigit() else "job_" + os.path.basename(progargs.jobfile) + "_" + str(pid)
    with open(progargs.jobfile) as fi:
        allcmds = fi.readlines()
        if progargs.endline is None:
            progargs.endline = len(allcmds)
        allcmds = allcmds[(progargs.startline-1):progargs.endline]
    allcmds = filter(not_empty, allcmds)
    progargs.alljobs = len(allcmds)
    if progargs.alljobs == 0:
        print("Error: No jobs for submitted!")
        sys.exit(os.EX_USAGE)
    if progargs.num is None:
        progargs.num = progargs.alljobs
    return progargs


def qsubCheck(jobname, num, block, sec=1):
    qs = 0
    global wait
    while True:
        time.sleep(sec)  # check per 1 seconds
        if not q.full():
            continue
        qs = os.popen('qstat -xml | grep %s | wc -l' % jobname).read().strip()
        qs = int(qs)
        if block or wait:
            if qs == 0:
                while True:
                    if q.empty():
                        wait = False
                        break
                    q.get()
            else:
                continue
        else:
            if qs < num:
                [q.get() for _ in range(num-qs)]
            else:
                continue


def checkAllSuccess(logdir):
    logfile = [os.path.join(logdir, i) for i in os.listdir(logdir)]
    stat = []
    for f in logfile:
        with open(f) as fi:
            ctx = fi.readlines()[-1].strip().split()[-1]
        if ctx == "SUCCESS":
            stat.append(True)
        else:
            stat.append(False)
    if len(stat) != 0 and all(stat):
        return True
    else:
        return False


def main():
    args = parserArg()
    jobfile = args.jobfile
    args.num = max(1, args.num)
    global q, jn, wait
    wait = False
    jn = args.jobname
    q = Queue(maxsize=args.num)
    p = Thread(target=qsubCheck, args=(args.jobname, args.num, args.block))
    p.setDaemon(True)
    p.start()
    h = ParseSingal()
    h.start()
    success = []
    with open(jobfile) as fi:
        if os.path.isdir(args.logdir):
            oldlog = [os.path.join(args.logdir, i)
                      for i in os.listdir(args.logdir)]
            for ol in oldlog:
                with open(ol) as oll:
                    ocmd = oll.readline().strip()
                    ostat = oll.readlines()[-1].strip().split()[-1]
                    if ocmd and ostat == "SUCCESS":
                        success.append(ocmd)
                    continue
            rmtree(args.logdir)
        os.makedirs(args.logdir)
        for n, line in enumerate(fi):
            line = line.strip().strip("& ")
            if n+1 < args.startline or n+1 > args.endline:
                continue
            if line.lower() == "wait":  # 若碰到wait, 则自动阻塞，等待前面所有任务执行完毕之后再开始投递
                wait = True
                while True:
                    if q.full():
                        break
                    q.put(n)
                continue
            if line and not line.startswith("#"):
                if line in success:
                    continue
                logfile = os.path.join(args.logdir, os.path.basename(
                    jobfile)+".line") + str(n+1) + ".log"
                # 300 hours  3600*300 = 1080000
                q.put(n, block=True, timeout=1080000)
                qsubline = line if line.endswith("ERROR") else line+RUNSTAT
                qsubline = "echo [\`date +'%F %T'\`] RUNNING... && " + qsubline
                cmd = 'qsub -q %s -wd %s -N %s -o %s -j y -l vf=%dg,p=%d <<< "%s"' % (args.queue,
                                                                                      args.workdir, args.jobname+"_" +
                                                                                      str(
                                                                                          n+1), logfile, args.memory, args.cpu, qsubline
                                                                                      )
                logcmd = open(logfile, "w")
                logcmd.write(line+"\n")
                logcmd.write("[%s] " % datetime.today().strftime("%F %X"))
                logcmd.flush()
                call(cmd, shell=True, stdout=logcmd, stderr=logcmd)
                logcmd.close()
    while True:
        time.sleep(2)  # check per 2 seconds
        qs = os.popen('qstat -xml | grep %s | wc -l' %
                      args.jobname).read().strip()
        qs = int(qs)
        if qs == 0:
            # '''
            # 此处可添加操作，所有任务完成之后进行的处理可写在这部分
            # '''
            qsubstat = checkAllSuccess(args.logdir)
            if qsubstat:
                print("[%s] All tesks in file (%s) finished successfully." % (
                    datetime.today().isoformat(), os.path.abspath(args.jobfile)))
                # rmtree(args.logdir)  # 删除logdir
            else:
                print("[%s] All tesks in file (%s) finished, But there are ERROR tesks." % (
                    datetime.today().isoformat(), os.path.abspath(args.jobfile)))
            break
    return


if __name__ == "__main__":
    main()



#!/usr/bin/env python2
# coding:utf-8

import os
import time
import random
import logging

from subprocess import call, PIPE, Popen
from collections import defaultdict, Counter
from threading import Thread
from Queue import Queue
from datetime import datetime

from job import Jobfile

RUNSTAT = " && echo [\`date +'%F %T'\`] SUCCESS || echo [\`date +'%F %T'\`] ERROR"


class myQueue(object):
    def __init__(self, maxsize=0):
        self._content = []
        self._queue = Queue(maxsize=maxsize)

    @property
    def length(self):
        return self._queue.qsize()

    def put(self, v, **kwargs):
        self._queue.put(v, **kwargs)
        self._content.append(v)

    def get(self, v=None):
        if v is None:
            o = self._content.pop(0)
            self._queue.get()
            return o
        else:
            if v not in self._content:
                return "%s not in queue" % v
            else:
                self._content.remove(v)
                self._queue.get()
                return v

    @property
    def queue(self):
        return self._content[:]

    def isEmpty(self):
        return self._queue.empty()

    def isFull(self):
        return self._queue.full()


class qsub(object):
    def __init__(self, jobfile, max_jobs=None, jobnames=None, start=1, end=None, mode=None, usestrict=False):
        self.pid = os.getpid()
        self.jfile = jobfile
        self.is_run = False
        self.firstjobnames = set()
        self.state = {}
        self.usestrict = usestrict

        jf = Jobfile(self.jfile, mode=mode)
        self.has_sge = jf.has_sge
        self.jobs = jf.jobs(jobnames, start, end)  # all jobs defined by args
        self.totaljobs = jf.totaljobs  # all jobs in job file
        self.totaljobdict = {jf.name: jf for jf in self.totaljobs}

        self.orders = jf.orders()  # total orders in job file
        self.orders_rev = {}   # total orders_rev in job file
        for k, v in self.orders.items():
            for i in v:
                self.orders_rev.setdefault(i, set()).add(k)

        # duplicate job names
        if len(jf.alljobnames) != len(set(jf.alljobnames)):
            names = [i for i, j in Counter(jf.alljobnames).items() if j > 1]
            self.throw("duplicate job name: %s" % " ".join(names))

        # order job name miss
        order_all = set(self.orders.keys() + self.orders_rev.keys())
        if order_all < jf.alljobnames:
            self.extrajob = jf.alljobnames - order_all
            self.throw("There are jobs not defined in orders")

        self.logdir = jf.logdir
        if not os.path.isdir(self.logdir):
            os.makedirs(self.logdir)

        self.error = set()  # args jobs error
        self.success = set()  # args jobs success,   self.error + self.success = len(self.jobs)
        self.thisjobnames = set([j.name for j in self.jobs])
        self.has_success = set()

        for jn in self.thisjobnames.copy():
            lf = os.path.join(self.logdir, jn + ".log")
            job = self.totaljobdict[jn]
            if job.status is not None and job.status in ["done", "success"]:
                self.thisjobnames.remove(jn)
                self.has_success.add(jn)
                self.success.add(jn)
                continue
            if os.path.isfile(lf):  # if log file not exists, will run
                js = self.jobstatus(jn)
                if js != "success":
                    os.remove(lf)
                    if js == "error":
                        # self.error.remove(jn)
                        self.state.pop(jn)
                else:
                    self.thisjobnames.remove(jn)  # thisjobs - has_success
                    self.has_success.add(jn)
                    self.success.add(jn)
            else:
                self.state[jn] = "wait"
        # thisjobnames are real jobs
        # len(self.has_success) + len(self.thisjobnames) = len(self.jobs)
        self.max_jobs = len(
            self.thisjobnames) if max_jobs is None else max_jobs

        self.jobqueue = myQueue(maxsize=self.max_jobs)

    def not_qsub(self, jobname):
        qs = os.popen('qstat -xml | grep %s_%d | wc -l' %
                      (jobname, self.pid)).read().strip()
        if int(qs) == 0:
            return True
        return False

    def jobstatus(self, jobname):
        js = self.state.get(jobname, "")
        if js == "success":
            return js
        elif js == "error":
            return js
        status = "wait"  # wait to submit
        logfile = os.path.join(self.logdir, jobname + ".log")
        if os.path.isfile(logfile):
            status = "submit"  # wait to run
            try:
                sta = os.popen('tail -n 1 %s' % logfile).read().split()[-1]
            except IndexError:
                status = "run"
                if js in ["submit", "resubmit"]:
                    self.logger.info("job %s status run", jobname)
                # if self.not_qsub(jobname) and self.is_run:                                                           ## job exit, qsub error
                #    self.throw("Error in %s job, probably because of qsub interruption."%jobname)
                self.state[jobname] = status
                return status
            if sta == "SUCCESS":
                status = "success"
                # self.success.add(jobname)
            elif sta == "ERROR":
                status = "error"
                # self.error.add(jobname)
            elif sta == "Exiting.":
                status = "exit"
            else:
                if "RUNNING..." in os.popen("sed -n '3p' %s" % logfile).read():
                    status = "run"
                    if js in ["submit", "resubmit"]:
                        self.logger.info("job %s status run", jobname)
                    # if self.not_qsub(jobname) and self.is_run:                                                       ## job exit, qsub error
                    #    self.throw("Error in %s job, probably because of qsub interruption."%jobname)
        self.state[jobname] = status
        return status

    def firstjob(self):
        queryjob_tmp = self.thisjobnames.copy()
        for j in self.thisjobnames:
            if j in self.orders:
                for bj in self.orders[j]:
                    if bj in self.thisjobnames:
                        queryjob_tmp.remove(j)
                        break
        self.firstjobnames.update(queryjob_tmp)
        return [self.totaljobdict[i] for i in self.firstjobnames]

    def jobcheck(self, sec=2):
        while True:
            time.sleep(sec/2)
            for jn in self.jobqueue.queue:
                time.sleep(sec/2)
                js = self.jobstatus(jn)
                if js in ["success", "error"]:
                    n = self.jobqueue.get(jn)

    def run(self, sec=2, times=-1, resubivs=2):

        self.is_run = True
        self.times = times
        self.subtimes = defaultdict(lambda: self.times)

        for jn in self.has_success:
            self.logger.info("job %s status already success", jn)

        firstqsub = self.firstjob()
        firstjobnames = set([j.name for j in firstqsub])
        if self.max_jobs < 100:
            p = Thread(target=self.jobcheck)
            p.setDaemon(True)
            p.start()
        prepare_sub = set()
        if len(firstqsub) <= self.max_jobs:
            for job in firstqsub:
                self.submit(job)
                if job.name in self.orders_rev:
                    # prepare_sub.update(
                    #    [i for i in self.orders_rev[job.name] if self.jobstatus(i) != "success"])
                    prepare_sub.update(
                        [i for i in self.orders_rev[job.name] if i in self.thisjobnames])
        else:
            prepare_sub = firstjobnames.copy()
            for j in prepare_sub:
                self.orders[j] = set([j])
        while len(self.thisjobnames) > 0:
            time.sleep(sec)
            for k in prepare_sub.copy():
                time.sleep(0.1)
                subK = True
                for jn in self.orders[k]:
                    time.sleep(0.1)
                    js = self.jobstatus(jn)
                    if js == "success":
                        if jn not in self.success:
                            self.logger.info("job %s status %s", jn, js)
                        if jn in self.error:
                            self.error.remove(jn)
                        self.success.add(jn)
                        continue
                    elif js == "error":
                        if jn not in self.error:
                            self.logger.info("job %s status %s", jn, js)
                        self.error.add(jn)
                        if self.subtimes[jn] < 0:
                            if self.usestrict:
                                self.throw("Error jobs return(resubmit %d times, still error), exist!, %s" % (self.times+1, os.path.join(
                                    self.logdir, jn + ".log")))  # if error, exit program
                            continue
                        else:
                            self.error.remove(jn)
                            time.sleep(resubivs)  # sleep, re-submit
                            self.submit(self.totaljobdict[jn], resub=True)
                            self.subtimes[jn] -= 1
                            subK = False
                    elif js == "exit":
                        self.throw("Error when qsub")
                    else:
                        if jn == k:
                            continue
                        subK = False
                if subK:
                    self.submit(self.totaljobdict[k])
                    if k in prepare_sub:
                        prepare_sub.remove(k)
                    if k in self.orders_rev:
                        prepare_sub.update(self.orders_rev[k])
                    for jn in self.orders[k]:
                        if jn in prepare_sub:
                            prepare_sub.remove(jn)
        self.finalstat(resubivs)

    def submit(self, job, resub=False):
        logfile = os.path.join(self.logdir, job.name + ".log")

        # if submit, jobstatus must not be "success", so don't need to do this condition.
        # if self.jobstatus(job.name) == "success":
        #    if job.name in self.thisjobnames:
        #        self.thisjobnames.remove(job.name)
        #    return
        if self.max_jobs < 100:
            self.jobqueue.put(job.name, block=True, timeout=1080000)

        if resub:
            logcmd = open(logfile, "a")
            logcmd.write("\n" + job.cmd+"\n")
            self.state[job.name] = "resubmit"
            self.logger.info("job %s status resubmit", job.name)
        else:
            logcmd = open(logfile, "w")
            logcmd.write(job.cmd+"\n")
            self.state[job.name] = "submit"
            self.logger.info("job %s status submit", job.name)
        logcmd.write("[%s] " % datetime.today().strftime("%F %X"))
        logcmd.flush()
        qsubline = "echo [\`date +'%F %T'\`] RUNNING... && " + \
            job.cmd + RUNSTAT
        if job.host is not None and job.host == "localhost":
            cmd = '''echo "Your job \\('%s'\\) has been submitted in localhost" && ''' % job.name + qsubline
            cmd = cmd.replace("\\", "")
            if resub:
                cmd = cmd.replace("RUNNING", "RUNNING \\(re-submit\\)")
            Popen(cmd, shell=True, stdout=logcmd, stderr=logcmd)
        else:
            job.cmd = job.cmd.replace('"', "\\\"")
            cmd = 'qsub %s -N %s_%d -o %s -j y <<< "%s"' % (
                job.sched_options, job.name, self.pid, logfile, qsubline)
            if resub:
                cmd = cmd.replace("RUNNING", "RUNNING \\(re-submit\\)")
            call(cmd, shell=True, stdout=logcmd, stderr=logcmd)
        logcmd.close()
        if job.name in self.thisjobnames:
            self.thisjobnames.remove(job.name)

    def finalstat(self, resubivs):
        finaljobs = set([j.name for j in self.jobs]) - self.has_success - \
            self.success - self.error
        while len(finaljobs) > 0:
            time.sleep(2)
            for jn in finaljobs.copy():
                js = self.jobstatus(jn)
                if js == "success":
                    finaljobs.remove(jn)
                    if jn not in self.success:
                        self.logger.info("job %s status %s", jn, js)
                    self.success.add(jn)
                    if jn in self.error:
                        self.error.remove(jn)
                elif js == "error":
                    if jn not in self.error:
                        self.logger.info("job %s status %s", jn, js)
                    self.error.add(jn)
                    if self.subtimes[jn] < 0:
                        finaljobs.remove(jn)
                        if self.usestrict:
                            self.throw("Error jobs return(resubmit %d times, still error), exist!, %s" % (self.times+1, os.path.join(
                                self.logdir, jn + ".log")))  # if error, exit program
                    else:
                        self.error.remove(jn)
                        time.sleep(resubivs)  # sleep, re-submit
                        self.submit(self.totaljobdict[jn], resub=True)
                        self.subtimes[jn] -= 1

    def throw(self, msg):
        raise RuntimeError(msg)

    def writejob(self, outjob):
        with open(outjob, "w") as fo:
            fo.write("log_dir " + self.logdir + "\n")
            for job in self.totaljobs:
                if job.name in self.state:
                    job.status = self.state[job.name]
                job.write(fo)
            for k, v in self.orders.items():
                for i in v:
                    fo.write("order %s after %s\n" % (k, i))

    def writestates(self, outstat):
        with open(outstat, "w") as fo:
            fo.write(str(dict(Counter(self.state.values()))) + "\n\n")
            sumary = {}
            for k, v in self.state.items():
                sumary.setdefault(v, []).append(k)
            for k, v in sorted(sumary.items(), key=lambda x: len(x[1])):
                fo.write(k + " : " + " ".join(v) + "\n")

    @property
    def logger(self):
        return logging.getLogger("__main__")
